require 'ap'
require 'pp'

require './drv_kdi.rb'


class CPM_RawDirEntry
	#http://www.seasip.demon.co.uk/Cpm/format22.html
	#Directory entries
	# The CP/M 2.2 directory has only one type of entry:

	# UU F1 F2 F3 F4 F5 F6 F7 F8 T1 T2 T3 EX S1 S2 RC   .FILENAMETYP....
	# AL AL AL AL AL AL AL AL AL AL AL AL AL AL AL AL   ................

	# UU = User number. 0-15 (on some systems, 0-31). The user number allows multiple
	#     files of the same name to coexist on the disc. 
	#      User number = 0E5h => File deleted
	# Fn - filename
	# Tn - filetype. The characters used for these are 7-bit ASCII.
	#        The top bit of T1 (often referred to as T1') is set if the file is 
	#      read-only.
	#        T2' is set if the file is a system file (this corresponds to "hidden" on 
	#      other systems). 
	# EX = Extent counter, low byte - takes values from 0-31
	# S2 = Extent counter, high byte.

	#       An extent is the portion of a file controlled by one directory entry.
	#     If a file takes up more blocks than can be listed in one directory entry,
	#     it is given multiple entries, distinguished by their EX and S2 bytes. The
	#     formula is: Entry number = ((32*S2)+EX) / (exm+1) where exm is the 
	#     extent mask value from the Disc Parameter Block.

	# S1 - reserved, set to 0.
	# RC - Number of records (1 record=128 bytes) used in this extent, low byte.
	#     The total number of records used in this extent is

	#     (EX & exm) * 128 + RC

	#     If RC is 80h, this extent is full and there may be another one on the disc.
	#     File lengths are only saved to the nearest 128 bytes.

	# AL - Allocation. Each AL is the number of a block on the disc. If an AL
	#     number is zero, that section of the file has no storage allocated to it
	#     (ie it does not exist). For example, a 3k file might have allocation 
	#     5,6,8,0,0.... - the first 1k is in block 5, the second in block 6, the 
	#     third in block 8.
	#      AL numbers can either be 8-bit (if there are fewer than 256 blocks on the
	#     disc) or 16-bit (stored low byte first). 
	#rawdata 0..31

	attr_accessor :status
	attr_accessor :name
	attr_accessor :ext
	attr_accessor :ex
	attr_accessor :s1
	attr_accessor :s2
	attr_accessor :rc
	attr_accessor :al


	def initialize(rawdata)
		@status,
		@name,
		@ext,
		@ex,
		@s1,
		@s2,
		@rc,
		*@al = rawdata[0..31].unpack("Ca8a3CCCCSSSSSSS")
		@name.force_encoding("koi8-r")
		@ext.force_encoding("koi8-r")
	end

	def to_s
		"#{@status} #{@name}.#{@ext} #{@ex} #{@s1} #{@s2} #{@rc}/#{@rc*128} [#{@al.join(",")}]"
	end
end

class DirEntry
	attr_accessor :filename
	attr_accessor :user
	attr_accessor :size_in_records
	attr_accessor :read_only
	attr_accessor :hidden
	attr_accessor :used_blocks

	def to_s
		r=read_only ? "R" : " "
		h=hidden    ? "H" : " "

		# flag_rus=filename.gsub(/[A-Za-z0-9\:\\\/\.\-@_\-\$\+]/,"")#.size > 0
		# "#{"%2d" % user}:#{r}#{h}:#{filename.ljust(13)}:>#{flag_rus.size>0}<:#{size_in_bytes}:>#{flag_rus}<:"

		"#{"%2d" % user}:#{r}#{h}:#{filename.ljust(13)}:#{size_in_bytes}"
	end

	def initialize(raw_dir_entry)
		ext=raw_dir_entry.ext
		@read_only=(ext[0].ord & 0x80)>0 ? true:false
		@hidden   =(ext[1].ord & 0x80)>0 ? true:false
		ext[0]=(ext[0].ord & 0x7f).chr
		ext[1]=(ext[1].ord & 0x7f).chr

		@filename="#{raw_dir_entry.name.rstrip}.#{ext.rstrip}".encode("utf-8")
		#@filename.encode("utf-8")
		#p [enc: @filename.encoding]

		@user=raw_dir_entry.status
		@used_blocks=[]
		#todo collect only rc related extends
		#@size_in_records=raw_dir_entry.rc
		#raw_dir_entry.al.each do |extent| 
	    #		@used_blocks << extent
		#en
		@size_in_records=0
		self.update(raw_dir_entry)
		self
	end

	def update(raw_dir_entry)
		@size_in_records+=raw_dir_entry.rc
		#todo collect only rc related extends
		raw_dir_entry.al.each do |extent| 
			@used_blocks << extent if extent>0
		end
		self
	end

	def size_in_bytes
		@size_in_records*128
	end
end


class CPM_Dir
	include Enumerable

	attr_accessor :dir

	def initialize(disk_drv)
		@disk_drv=disk_drv

		@raw_dir=getFileList
		@dir = []


		#extract all files with current ex=0  (and not erased)
		files_0=@raw_dir.select{|v|
			r=true
			r=false unless v.status != 0xe5 		#erased
			r=false unless v.ex == 0 and v.s2 == 0  #extend 0 (and e)
			r
		}

		files_0.each do |raw_dir_entry|
			#puts raw_dir_entry
			file=DirEntry.new(raw_dir_entry)
			cc=raw_dir_entry
			while cc=get_next(cc)
				#p  [:while,cc]
				file.update(cc)
			end
			@dir << file
		end
	end


	def each &block
	    @dir.each{|member| block.call(member)}
	end	

	private

	#read raw directory
	def getFileList
		raw_dir=getRawDirectory
		dir=[]
		@disk_drv.dpb.drm.times do |n|
			#p [n,32*n...(32*(n+1))-1]
			dir << CPM_RawDirEntry.new(raw_dir[32*n...(32*(n+1))-1])
		end
		dir
	end

	def getRawDirectory
		dirblock=(@disk_drv.dpb.al1 + @disk_drv.dpb.al0<<8).to_s(2).gsub("0","").size
		raw_dir=""
		0.upto(dirblock-1) do |b|
			raw_dir+=@disk_drv.get_cpm_block(b)
		end
		raw_dir
	end

	def get_next(current_extent)
		xnext=current_extent.clone
		if xnext.ex == 31
			xnext.s2+=1
			xnext.ex=0
		else
			xnext.ex+=1
		end

		#p [:try_next,xnext]
		ne=@raw_dir.select{|v|
			r=false
			r=true if 	v.status == xnext.status and
						v.name   == xnext.name   and
						v.ext    == xnext.ext    and
						v.ex     == xnext.ex     and 
						v.s2     == xnext.s2
			r			
		}
		r=nil
		r=ne[0] if ne.size>=1
		#p [:gn,ne.size,r]
		r
	end
end	


class CPM_Drive

	def initialize(disk_drv)
		@disk_drv=disk_drv
		@dir=CPM_Dir.new(disk_drv)
	end

	def get_dir
		@dir
	end

	def get_dpb
		@disk_drv.get_dpb
	end

	def get_file(dir_entry)
		p [:blocks,dir_entry.filename,dir_entry.size_in_bytes,dir_entry.used_blocks]
		body=""
		dir_entry.used_blocks.each do |block|
			work_block=@disk_drv.get_cpm_block(block)
			body+=work_block
		end
		body[0...dir_entry.size_in_bytes]
	end

	def check_disk
		error_flag=false

		blocks_on_disk=@disk_drv.get_dpb.dsm
		blocks_map=Array.new(blocks_on_disk)

		@dir.each do |dir_entry|
			dir_entry.used_blocks.each do |block|
				if block>blocks_on_disk
					p [:chkdsk,:block_to_big,dir_entry.filename,block]
					error_flag=true
				else
					if owner=blocks_map[block]
						p [:chkdsk,:block_used,dir_entry.filename,block,owner.filename]
						error_flag=true
					else
						blocks_map[block]=dir_entry
					end
				end
			end
		end
		error_flag
	end

end





if __FILE__ == $0 #PROGRAM_NAME
	file_out=File.new("out.txt","w")
	#_disk_drv=KDI.new("./KDI/disk1.kdi")
	#_disk_drv=KDI.new("./KDI/disk2.kdi")
	#_disk_drv=KDI.new("./KDI/disk3.kdi")
	#_disk_drv=KDI.new("./KDI/disk4.kdi")
	#_disk_drv=KDI.new("./KDI/AG_dict_en-ru.kdi")
	list=Dir["/home/serokhin/Dropbox/Emulator/Korvet/KorvetSoft/**/*.kdi"]
	list+=Dir["/home/serokhin/Dropbox/Emulator/Korvet/KorvetSoft/**/*.KDI"]
	# list=Dir["/Users/esl/Dropbox/Emulator/Korvet/KorvetSoft/**/*.kdi"]
	# list+=Dir["/Users/esl/Dropbox/Emulator/Korvet/KorvetSoft/**/*.KDI"]
	list=["/home/serokhin/Dropbox/Emulator/Korvet/KorvetSoft/Shtirlitz/DE_6_01.kdi"]
	#list=["/home/serokhin/Dropbox/Emulator/Korvet/KorvetSoft/AlexanderGrigoryev/AG_kornet_drive-a_gamz.kdi"]

	disk_with_error=0
	list=list.sort
	list.each do |name|
		p [:name,name]
		file_out.puts "[#{name}]"
		begin

			_disk_drv=KDI.new(name)
			disk=CPM_Drive.new(_disk_drv)
			#puts disk.dir.get_dir
			file_out.puts disk.get_dir

			# fnames=disk.get_dir.map{|v| user=v.user>0?"\t\t\t#{v.user}:" : ""; "#{user}#{v.filename}"}
			# puts fnames


			#disk_with_error+=1 if disk.check_disk

			# name2=name.clone.downcase
			# name2.gsub!(/.*Korvet\/KorvetSoft\//i,"")
			# name2.gsub!(/\//,"_")

			# p [:name2,name2]
			p [:ofs,disk.get_dpb.ofs]
			p [:spt,disk.get_dpb.spt*128]
			p [:blm,128*(disk.get_dpb.blm+1)]

			# disk.get_dir.each do |entry|
			# 	p entry.used_blocks
			# 	# File.open("1/#{entry.filename}", "w") { |io| io.puts disk.get_file(entry)  }
			# end
		# rescue Exception => e
		# 	puts "Exception: " 
		# 	puts e#.message  
		end
	end
	# p [:disk,list.size]
	# p [:with_error,disk_with_error]
	#xd=disk.d2
	#p xd
	# 0.upto(127-1) { |n|
	# 	p [n,xd[n]] if xd[n] #unless disk.raw_dir.dir[n].status == 0xe5
	# }
end


