require 'ap'
require 'pp'
#KDI (Korvet Disk Image) disk support


class CPM_DirEntry
	#http://www.seasip.demon.co.uk/Cpm/format22.html
	#Directory entries
	# The CP/M 2.2 directory has only one type of entry:

	# UU F1 F2 F3 F4 F5 F6 F7 F8 T1 T2 T3 EX S1 S2 RC   .FILENAMETYP....
	# AL AL AL AL AL AL AL AL AL AL AL AL AL AL AL AL   ................

	# UU = User number. 0-15 (on some systems, 0-31). The user number allows multiple
	#     files of the same name to coexist on the disc. 
	#      User number = 0E5h => File deleted
	# Fn - filename
	# Tn - filetype. The characters used for these are 7-bit ASCII.
	#        The top bit of T1 (often referred to as T1') is set if the file is 
	#      read-only.
	#        T2' is set if the file is a system file (this corresponds to "hidden" on 
	#      other systems). 
	# EX = Extent counter, low byte - takes values from 0-31
	# S2 = Extent counter, high byte.

	#       An extent is the portion of a file controlled by one directory entry.
	#     If a file takes up more blocks than can be listed in one directory entry,
	#     it is given multiple entries, distinguished by their EX and S2 bytes. The
	#     formula is: Entry number = ((32*S2)+EX) / (exm+1) where exm is the 
	#     extent mask value from the Disc Parameter Block.

	# S1 - reserved, set to 0.
	# RC - Number of records (1 record=128 bytes) used in this extent, low byte.
	#     The total number of records used in this extent is

	#     (EX & exm) * 128 + RC

	#     If RC is 80h, this extent is full and there may be another one on the disc.
	#     File lengths are only saved to the nearest 128 bytes.

	# AL - Allocation. Each AL is the number of a block on the disc. If an AL
	#     number is zero, that section of the file has no storage allocated to it
	#     (ie it does not exist). For example, a 3k file might have allocation 
	#     5,6,8,0,0.... - the first 1k is in block 5, the second in block 6, the 
	#     third in block 8.
	#      AL numbers can either be 8-bit (if there are fewer than 256 blocks on the
	#     disc) or 16-bit (stored low byte first). 
	#rawdata 0..31
	attr_reader :status
	attr_reader :name
	attr_reader :ext
	attr_reader :ex
	attr_reader :s1
	attr_reader :s2
	attr_reader :rc
	attr_reader :al


	def initialize(rawdata)
		@status,
		@name,
		@ext,
		@ex,
		@s1,
		@s2,
		@rc,
		*@al = rawdata[0..31].unpack("Ca8a3CCCCSSSSSSS")

		# p [
		# 	@status,
		# 	@name,
		# 	@ext,
		# 	@ex,
		# 	@s1,
		# 	@s2,
		# 	@rc,@rc*128,
		# 	@al
		# 	]
		#p [to_s]
	end

	def to_s
		"#{@status} #{@name}.#{@ext} #{@ex} #{@s1} #{@s2} #{@rc}/#{@rc*128} [#{@al.join(",")}]"
	end
end


class KDIDiskImage

	#physical level
	@sector_size=nil
	@tracks_per_disk=nil
	@sector_per_track=nil

	@block0_offset=nil

	#logical level
	@system_tracks=nil
	@block_size=nil	
	@dir_entrirs=nil

	def initialize(filename)
		@source_file=filename
		@kdi_image=File.open(filename,"rb").read
#		p @kdi_image.size
		#raise "Too small image [#{filename}]" unless @kdi_image.size > 128

		calculated_crc=@kdi_image[0..31-1].
						split('').
						reduce(0x66){|sum,n| (sum+n.ord)&0xff}


		readDiskParam()


		raise "Invalid Disk CRC [#{filename}] " unless calculated_crc == @CheckSum

		#check logical disk params
		raise "BLS error, check BSH,BLM [#{filename}]" unless 128*(1<<@BSH) == 128*(@BLM+1)

		@sector_size=0x80<<@SecSize

		@block_size=128*(1<<@BSH)
		@system_tracks=@OFS
		@dir_entrirs=@DRM

		@block0_offset=@Count*@sector_size

		# p [:bls,@BLS]
		# p [:AL,(@AL1 + @AL0<<8).to_s(2)]
	end

	def to_s
		"#{"%6d" % @kdi_image.size} [#{@sector_size},#{@block_size},#{@system_tracks},#{@dir_entrirs},#{@SPT}](#{@Count}) #{@source_file.ljust(40)}"
	end

	def getBlock(block_number)
		b0=@block0_offset+(@block_size*block_number)
		b1=b0+@block_size
		@kdi_image[b0...b1]
	end

	def getRawSystem
		if @Count>0
			@kdi_image[0...(@Count*@sector_size)]
		else
			""
		end
	end

	def getRawDirectory
		dirblock=(@AL1 + @AL0<<8).to_s(2).gsub("0","").size
		raw_dir=""
		0.upto(dirblock-1) do |b|
			raw_dir+=getBlock(b)
		end
		raw_dir
	end

	def GetFileList
		raw_dir=getRawDirectory
		dir=[]
		@DRM.times do |n|
			#p [n,32*n...(32*(n+1))-1]
			dir << CPM_DirEntry.new(raw_dir[32*n...(32*(n+1))-1])
		end
		dir
	end

	def GetFile
	end

	def readDiskParam

		# ;Eфизические параметры дискаF

		#  1:    DW LoadAdress ; адрес, с которого начинается загрузка
		#  3:    DW RunAdress  ; адрес,куда передается управление после загрузки
		#  5:    DW Count      ; число загружаемых физических секторов
		#  7:    DB SizeDisk   ; если значение байта 1 -8" диск, если 0 - 5.25"
		#  8:    DB Density    ; способ записи: 0 - FM; 1 - MFM
		#  9:    DB TpI        ; число дорожек  на дюйм: 0 - 48 TpI 1 - 96 TpI, 2 - 135 TpI
		# 10:    DB SkewFactor ; если значение байта 1 - фактора смещения
		#                      ; нет. Если <> 1 - байты 33-128
		#                      ; информационного сектора содержат таблицу
		#                      ; перевода секторов, причем значение байта
		#                      ; равно размеру таблицы.
		# 11:    DB SecSize    ; размер физического сектора:0 - 128 байт, 1 - 256 байт, 2 - 512 байт и  3 - 1024
		# 12:    DB InSide     ; информация о сторонах диска: 
		#					   ; 0 - односторонний диск, 1 - двухсторонний,
		#                      ; причем сектора нумеруются от 1 до n 
		#                      ; с каждой стороны, четные логические
		#                      ; дорожки с нулевой стороны, нечетные с
		#                      ; первой.
		# 13:    DW SecPerTrack; число физических секторов на дорожке
		# 15:    DW TrkPerDisk ; число дорожек на диске (с одной стороны)

		# ;Eлогические параметры дискаF

		# 17:    DW SPT        ; число логических секторов на дорожке
		# 19:    DB BSH        ; фактор сдвига блока распределения данных
		# 20:    DB BLM        ; маска блока распределения данных
		# 21:    DB EXM        ; маска размера
		# 22:    DW DSM        ; объем памяти на диске в блоках минус 1
		# 24:    DW DRM        ; число входов в директорию диска минус 1
		# 26:    DB AL0        ; определяет, какие блоки зарезервированы
		# 27:    DB AL1        ; под директорию
		# 28:    DW CKS        ; размер вектора контроля директории
		# 30:    DW OFS        ; число системных дорожек на диске
		# 32:    DB CheckSum   ;контрольная сумма служебной информации
		#                                                      CS[1-31]+66H

		@LoadAdress ,
		@RunAdress  ,
		@Count      ,
		@SizeDisk   ,
		@Density    ,
		@TpI        ,
		@SkewFactor ,
		@SecSize    ,
		@InSide     ,
		@SecPerTrack,
		@TrkPerDisk ,

		@SPT        ,		# No. of Logical ( 128-byte ) Sectors per Logical Track
		@BSH        ,		# Block Shift - Block Size is given by 128 * 2^(BSH)
		@BLM        ,		# Block Mask - Block Size is given by 128 * (BLM +1)
		@EXM        ,		# Extent Mask ( see opposite )
		@DSM        ,		# объем памяти на диске в блоках минус 1
		@DRM        ,		# число входов в директорию диска минус 1
		@AL0        ,		# определяет, какие блоки зарезервированы
		@AL1        ,		# под директорию
		@CKS        ,		# размер вектора контроля директории
		@OFS        ,		# число системных дорожек на диске
		@CheckSum 			= @kdi_image[0..31].unpack("SSSCCCCCCSS"+"SCCCSSCCSSC")


		#http://sharpmz.org/dpb.htm
		#http://sharpmz.org/succpminfo06.htm

		#BSH
		#Data allocation "Block Shift Factor"
		#Number of 128 bytes sectors per "Allocation Block" 
		#( determined by the data block allocation size. Stored as 2's-logarithm. 
		#See notes below ).

		#BLM
		#"Data allocation Block Mask"
		#To get the sector index of the Allocation Block the sector number is to 
		#combine logically by AND with this mask. (( 2^BSH)-1). See notes below.


		#The values of BSH and BLM implicitly determine the data allocation size, BLS,
		#which is not an entry in the DPB. 
		#Given that the designer has selected a value for BLS, 
		#the values of BSH and BLM are shown in the following table:

		#BLS 	BSH	BLM
		# 1,024	  3   7
		# 2,048	  4  15
		# 4,096   5  31
		# 8,192   6  63
		#16,384   7 127


		#EXM	"Extent Mask"
		#Number of extents per directory entry ( stored as 2's logarithm ). 
		#Determined by the data block allocation size and the number of disk blocks.

		#DSM	
		#Total storage capacity of the disk drive. 
		#Number of the last Allocation Block ( Number of entries per disk -1 )

		#DRM	
		#Number of the last directory entry ( Number of entries per disk -1 ). 
		#Total number of directory entries that can be stored on this drive.
		#( AL0, AL1 determine reserved directory blocks ).

		#AL0& AL1
		#
		#These parameters, taken BITWISE, indicate which of the first 16 CP/M blocks ( numbered 00 - 15 ) are allocated to the Directory. For example:-
		#$80 00 = 10000000 00000000 i.e. the Directory is allocated BLOCK 00 only
		#$C0 00= 11000000 00000000 i.e. the Directory is allocated BLOCKS 00 & 01

		#CKS "Check area Size"
		#Size of the directory check vector. 
		#Number of directory entries to check for disk change. Zero for a HDD.

		#OFF	
		#Number of system reserved tracks at the beginning of the ( logical ) disk


		# p [@LoadAdress ,@RunAdress].map{|v| "%04x" % v}
		# pp [
		# [:Count      ,@Count      ],
		# [:SizeDisk   ,@SizeDisk   ],
		# [:Density    ,@Density    ],
		# [:TpI        ,@TpI        ],
		# [:SkewFactor ,@SkewFactor ],
		# [:SecSize    ,@SecSize    ],
		# [:InSide     ,@InSide     ],
		# [:SecPerTrack,@SecPerTrack],
		# [:TrkPerDisk ,@TrkPerDisk ],

		# nil,

		# [:SPT,@SPT],
		# [:BSH,@BSH],
		# [:BLM,@BLM],
		# [:EXM,@EXM],
		# [:DSM,@DSM],
		# [:DRM,@DRM],
		# [:AL0,@AL0],
		# [:AL1,@AL1],
		# [:CKS,@CKS],
		# [:OFS,@OFS],
		# @CheckSum	
		# ]#.map{ |v| "%4x" % v }
	end

end

if __FILE__ == $0 #PROGRAM_NAME
	#a=KDIDiskImage.new("./KDI/disk1.kdi")
	#list=Dir["./KDI/*.kdi"]
	list=Dir["./os/*"]
	#list=Dir["/home/serokhin/Dropbox/Emulator/Korvet/KorvetSoft/*/*.kdi"]
	#list+=Dir["/home/serokhin/Dropbox/Emulator/Korvet/KorvetSoft/*/*.KDI"]
	p list.size

	h_dsize={}

	list.each do |n|
		kdi=KDIDiskImage.new(n)
		dst=n.gsub("/Users/esl/Dropbox/Emulator/Korvet/KorvetSoft/","")
		dst.gsub!("/","_")
		dst.gsub!(".","_")
		p kdi
	#	File.open("out/#{dst}","wb") do |f| f.write(kdi.getRawSystem) end
		#File.open("out/#{dst}","wb") do |f| f.write(kdi.getRawDirectory) end

		block_list=[]

		system_size=kdi.getRawSystem.size

		p [:s0,n] if system_size==0x3400

		h_dsize["%04x" % system_size]=n

		# kdi.GetFileList.each do |l|

		# 	if l.status != 229
		# 		#p l
		# 		#p l.al
		# 		block_list+=l.al
		# 		block_list.delete_if {|v| v == 0}
		# 	end
		# end

		# block_list=block_list.sort
		#dup_list=detect_dup(block_list)
		#p block_list
		# if dup_list.size > 0
		# 	puts "ERROR :#{dup_list.size}" 
		# 	p [dup_list]
		# end

	end
end

ap h_dsize